@@ -10,29 +10,42 @@
  * See the file LICENSE for distribution terms.
  */
 
+#ifdef USE_OMCACHE
+#define OMCACHE_READ_TIMEOUT 2000
+#include "omcache_libmemcached.h"
+#include <syslog.h>  /* for log levels */
+#endif /* USE_OMCACHE */
+#ifdef USE_LIBMEMCACHED
+#include <libmemcached/memcached.h>
+#endif /* USE_LIBMEMCACHED */
+
 #include "pgmemcache.h"
 
+#define KEY_MAX_LENGTH 250
+
 #ifdef PG_MODULE_MAGIC
 PG_MODULE_MAGIC;
 #endif
 
 /* Internal functions */
 static void pgmemcache_reset_context(void);
+static void pgmemcache_xact_callback(XactEvent event, void *arg);
 static void assign_sasl_params(const char *username, const char *password);
 static void assign_default_servers_guc(const char *newval, void *extra);
 static void assign_default_behavior_guc(const char *newval, void *extra);
 static memcached_behavior get_memcached_behavior_flag(const char *flag);
-static uint64_t get_memcached_behavior_data(const char *flag, const char *data);
-static uint64_t get_memcached_hash_type(const char *data);
-static uint64_t get_memcached_distribution_type(const char *data);
-static Datum memcache_atomic_op(bool increment, PG_FUNCTION_ARGS);
+static uint64_t get_memcached_behavior_data(const char *flag, const char *data, const char **val);
 static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS);
 static memcached_return do_server_add(const char *host_str);
+const char *get_arg_cstring(text *text_field, size_t *length, bool is_key);
+
 
 /* Per-backend global state. */
 static struct memcache_global_s
 {
   memcached_st *mc;
+  bool flush_needed;
+  bool flush_on_commit;
   char *default_servers;
   char *default_behavior;
   char *sasl_authentication_username;
@@ -70,6 +83,19 @@ void _PG_init(void)
                              assign_default_behavior_guc,
                              NULL);
 
+  DefineCustomBoolVariable("pgmemcache.flush_on_commit",
+                           "Whether to flush all buffers to memcached on end of commit",
+                           NULL,
+                           &globals.flush_on_commit,
+                           false,
+                           PGC_USERSET,
+                           0,
+#if defined(PG_VERSION_NUM) && (PG_VERSION_NUM >= 90100)
+                           NULL,
+#endif
+                           NULL,
+                           NULL);
+
   DefineCustomStringVariable("pgmemcache.sasl_authentication_username",
                              "pgmemcache SASL user authentication username",
                              "Simple string pgmemcache.sasl_authentication_username = 'testing_username'",
@@ -99,6 +125,8 @@ void _PG_init(void)
   /* XXX: We should have real assign_hook for these so that memcache sasl data
    * is updated when the values are changed.  */
   assign_sasl_params(globals.sasl_authentication_username, globals.sasl_authentication_password);
+
+  RegisterXactCallback(pgmemcache_xact_callback, NULL);
 }
 
 /* This is called when we're being unloaded from a process. Note that
@@ -131,10 +159,46 @@ static time_t interval_to_time_t(Interval *span)
   return (time_t) result;
 }
 
+/* called at end of transaction, flush all buffers to memcache */
+static void pgmemcache_xact_callback(XactEvent event, void *arg)
+{
+  if (globals.flush_on_commit && globals.flush_needed &&
+      (event == XACT_EVENT_COMMIT
+#if defined(PG_VERSION_NUM) && (PG_VERSION_NUM >= 90300)
+      || event == XACT_EVENT_PRE_COMMIT
+#endif /* PG_VERSION_NUM >= 90300 */
+      ))
+    {
+#ifdef USE_LIBMEMCACHED
+      memcached_return rc = memcached_flush_buffers(globals.mc);
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+      int rc = omcache_io(globals.mc, NULL, NULL, NULL, NULL, OMCACHE_READ_TIMEOUT);
+#endif /* USE_OMCACHE */
+      if (rc != MEMCACHED_SUCCESS)
+        elog(WARNING, "pgmemcache: memcached_flush_buffers: %s",
+                      memcached_strerror(globals.mc, rc));
+      else
+        globals.flush_needed = false;
+    }
+}
+
+#ifdef USE_OMCACHE
+static void pgmemcache_log_func(void *context, int syslog_level, const char *msg)
+{
+  int pg_level = LOG_INFO;
+  switch (syslog_level)
+    {
+    case LOG_ERR:
+    case LOG_WARNING: pg_level = WARNING; break;
+    case LOG_NOTICE: pg_level = NOTICE; break;
+    }
+  elog(pg_level, "%s", msg);
+}
+#endif /* USE_OMCACHE */
+
 static void pgmemcache_reset_context(void)
 {
-  int rc;
-
   if (globals.mc)
     {
       memcached_free(globals.mc);
@@ -143,14 +207,22 @@ static void pgmemcache_reset_context(void)
 
   globals.mc = memcached_create(NULL);
 
+#ifdef USE_OMCACHE
+  omcache_set_log_callback(globals.mc, 0, pgmemcache_log_func, NULL);
+#endif /* USE_OMCACHE */
+
+#ifdef USE_LIBMEMCACHED
   /* Always use the memcache binary protocol as required for
      memcached_(increment|decrement)_with_initial. */
-  rc = memcached_behavior_set(globals.mc,
-                              MEMCACHED_BEHAVIOR_BINARY_PROTOCOL,
-                              1);
-  if (rc != MEMCACHED_SUCCESS)
-    elog(WARNING, "pgmemcache: memcached_behavior_set(BINARY_PROTOCOL): %s",
-                  memcached_strerror(globals.mc, rc));
+  {
+    int rc = memcached_behavior_set(globals.mc,
+                                    MEMCACHED_BEHAVIOR_BINARY_PROTOCOL,
+                                    1);
+    if (rc != MEMCACHED_SUCCESS)
+      elog(WARNING, "pgmemcache: memcached_behavior_set(BINARY_PROTOCOL, 1): %s",
+                    memcached_strerror(globals.mc, rc));
+  }
+#endif /* USE_LIBMEMCACHED */
 
   assign_default_behavior_guc(globals.default_behavior, NULL);
   assign_sasl_params(globals.sasl_authentication_username, globals.sasl_authentication_password);
@@ -176,9 +248,11 @@ static void assign_default_servers_guc(const char *newval, void *extra)
 {
   if (newval)
     {
+#ifdef USE_LIBMEMCACHED
       /* there is no way to remove servers from a memcache context, so
        * recreate it from scratch when the server list changes */
       pgmemcache_reset_context();
+#endif /* USE_LIBMEMCACHED */
       do_server_add(newval);
     }
 }
@@ -189,6 +263,10 @@ static void assign_default_behavior_guc(const char *newval, void *extra)
   StringInfoData flag_buf;
   StringInfoData data_buf;
   memcached_return rc;
+  memcached_behavior bkey;
+  uint64_t bval;
+  const char *bvalstr = "";
+
   if (!newval)
     return;
 
@@ -221,11 +299,72 @@ static void assign_default_behavior_guc(const char *newval, void *extra)
 
               i += data_buf.len;
             }
-          rc = memcached_behavior_set(globals.mc,
-                                      get_memcached_behavior_flag(
-                                        flag_buf.data),
-                                      get_memcached_behavior_data(
-                                        flag_buf.data, data_buf.data));
+          rc = MEMCACHED_FAILURE;
+          bkey = get_memcached_behavior_flag(flag_buf.data);
+          bval = get_memcached_behavior_data(flag_buf.data, data_buf.data, &bvalstr);
+#ifdef USE_OMCACHE
+          if (bkey == NULL)
+            rc = OMCACHE_FAIL;
+          else if (strcmp(bkey, "BINARY_PROTOCOL") == 0)
+            {
+              if (!bval)
+                elog(ERROR, "pgmemcache: omcache always uses binary protocol");
+              rc = OMCACHE_OK;
+            }
+          else if (strcmp(bkey, "BUFFER_REQUESTS") == 0)
+            rc = omcache_set_buffering(globals.mc, bval);
+          else if (strcmp(bkey, "CONNECT_TIMEOUT") == 0)
+            rc = omcache_set_connect_timeout(globals.mc, bval);
+          else if (strcmp(bkey, "DEAD_TIMEOUT") == 0)
+            rc = omcache_set_dead_timeout(globals.mc, bval * 1000);
+          else if (strcmp(bkey, "DISTRIBUTION") == 0)
+            {
+              if (strcmp(bvalstr, "CONSISTENT") && strcmp(bvalstr, "CONSISTENT_KETAMA"))
+                elog(ERROR, "pgmemcache: omcache always uses ketama");
+              rc = OMCACHE_OK;
+            }
+          else if (strcmp(bkey, "HASH") == 0 || strcmp(bkey, "KETAMA_HASH") == 0)
+            {
+              if (strcmp(bvalstr, "DEFAULT"))
+                elog(ERROR, "pgmemcache: omcache always uses the 'default' (bob jenkins' 'one at a time') hash");
+              rc = OMCACHE_OK;
+            }
+          else if (strcmp(bkey, "KETAMA") == 0)
+            {
+              if (!bval)
+                elog(ERROR, "pgmemcache: omcache always uses a ketama distribution method");
+              rc = omcache_set_distribution_method(globals.mc, &omcache_dist_libmemcached_ketama);
+            }
+          else if (strcmp(bkey, "KETAMA_WEIGHTED") == 0)
+            {
+              if (!bval)
+                elog(ERROR, "pgmemcache: omcache always uses a ketama distribution method");
+              rc = omcache_set_distribution_method(globals.mc, &omcache_dist_libmemcached_ketama_weighted);
+            }
+          else if (strcmp(bkey, "KETAMA_PRE1010") == 0)
+            {
+              if (!bval)
+                elog(ERROR, "pgmemcache: omcache always uses a ketama distribution method");
+              rc = omcache_set_distribution_method(globals.mc, &omcache_dist_libmemcached_ketama_pre1010);
+            }
+          else if (strcmp(bkey, "NO_BLOCK") == 0)
+            rc = OMCACHE_OK;  // omcache is non-blocking by default
+          else if (strcmp(bkey, "NOREPLY") == 0)
+            rc = OMCACHE_OK;  // this isn't really a behavior in omcache
+          else if (strcmp(bkey, "REMOVE_FAILED_SERVERS") == 0)
+            rc = OMCACHE_OK;  // omcache doesn't have this concept
+          else if (strcmp(bkey, "RETRY_TIMEOUT") == 0)
+            rc = omcache_set_reconnect_timeout(globals.mc, bval * 1000);
+          else if (strcmp(bkey, "SUPPORT_CAS") == 0)
+            rc = OMCACHE_OK;  // omcache uses binary protocol which always has cas
+          else
+            {
+              elog(ERROR, "pgmemcache: unsupported behavior %s for omcache", flag_buf.data);
+            }
+#endif /* USE_OMCACHE */
+#ifdef USE_LIBMEMCACHED
+          rc = memcached_behavior_set(globals.mc, bkey, bval);
+#endif /* USE_LIBMEMCACHED */
           if (rc != MEMCACHED_SUCCESS)
             elog(WARNING, "pgmemcache: memcached_behavior_set: %s",
                           memcached_strerror(globals.mc, rc));
@@ -247,37 +386,43 @@ static void assign_default_behavior_guc(const char *newval, void *extra)
 
 Datum memcache_add(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_ADD | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_ADD | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
 }
 
 Datum memcache_add_absexpire(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_ADD | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_ADD | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
 }
 
-static Datum memcache_atomic_op(bool increment, PG_FUNCTION_ARGS)
+const char *get_arg_cstring(text *text_field, size_t *length, bool is_key)
 {
-  text *atomic_key = PG_GETARG_TEXT_P(0);
-  char *key;
-  size_t key_length;
-  uint64_t val;
-  uint64_t offset = 1;
-  memcached_return rc;
-
-  key = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(atomic_key)));
-  key_length = strlen(key);
-
-  if (key_length < 1)
+  *length = VARSIZE(text_field) - VARHDRSZ;
+  if (is_key && *length < 1)
     elog(ERROR, "pgmemcache: key cannot be an empty string");
-  if (key_length >= 250)
-    elog(ERROR, "pgmemcache: key too long");
+  else if (is_key && *length > KEY_MAX_LENGTH)
+    elog(ERROR, "pgmemcache: key too long, maximum is %d characters", KEY_MAX_LENGTH);
+  return VARDATA(text_field);
+}
+
+static Datum memcache_delta_op(bool increment, PG_FUNCTION_ARGS)
+{
+  uint64_t val;
+  int64_t offset = 1;
+  memcached_return rc;
+  size_t key_length;
+  const char *key = get_arg_cstring(PG_GETARG_TEXT_P(0), &key_length, true);
 
   if (PG_NARGS() >= 2)
     offset = PG_GETARG_INT64(1);
+
   if (offset < 0)
     {
-      /* Cannot represent negative BIGINT values with uint64_t */
-      elog(ERROR, "pgmemcache: offset cannot be negative");
+      /* memcached uses uint64_t but postgresql only has signed types, but
+       * since we have both increment and decrement operations let's just
+       * invert the operation if offset is negative.
+       */
+      offset = abs(offset);
+      increment = !increment;
     }
 
   if (increment)
@@ -285,6 +430,11 @@ static Datum memcache_atomic_op(bool increment, PG_FUNCTION_ARGS)
   else
     rc = memcached_decrement_with_initial(globals.mc, key, key_length, offset, 0, MEMCACHED_EXPIRATION_NOT_ADD, &val);
 
+  if (rc == MEMCACHED_BUFFERED)
+    {
+      globals.flush_needed = true;
+      PG_RETURN_NULL();
+    }
   if (rc != MEMCACHED_SUCCESS)
     {
       elog(WARNING, "pgmemcache: memcached_%s_with_initial: %s",
@@ -305,35 +455,36 @@ static Datum memcache_atomic_op(bool increment, PG_FUNCTION_ARGS)
 
 Datum memcache_decr(PG_FUNCTION_ARGS)
 {
-  return memcache_atomic_op(false, fcinfo);
+  return memcache_delta_op(false, fcinfo);
+}
+
+Datum memcache_incr(PG_FUNCTION_ARGS)
+{
+  return memcache_delta_op(true, fcinfo);
 }
 
 Datum memcache_delete(PG_FUNCTION_ARGS)
 {
-  text *key_to_be_deleted = PG_GETARG_TEXT_P(0);
-  size_t key_length;
   time_t hold;
   memcached_return rc;
-  char *key;
-
-  key = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(key_to_be_deleted)));
-  key_length = strlen(key);
-  if (key_length < 1)
-    elog(ERROR, "pgmemcache: key cannot be an empty string");
-  if (key_length >= 250)
-    elog(ERROR, "pgmemcache: key too long");
+  size_t key_length;
+  const char *key = get_arg_cstring(PG_GETARG_TEXT_P(0), &key_length, true);
 
   hold = (time_t) 0.0;
   if (PG_NARGS() >= 2 && PG_ARGISNULL(1) == false)
     hold = interval_to_time_t(PG_GETARG_INTERVAL_P(1));
 
   rc = memcached_delete(globals.mc, key, key_length, hold);
-
+  if (rc == MEMCACHED_BUFFERED)
+    {
+      globals.flush_needed = true;
+      PG_RETURN_NULL();
+    }
   if (rc != MEMCACHED_SUCCESS && rc != MEMCACHED_NOTFOUND)
     elog(WARNING, "pgmemcache: memcached_delete: %s",
                   memcached_strerror(globals.mc, rc));
 
-  PG_RETURN_BOOL(rc == 0);
+  PG_RETURN_BOOL(rc == MEMCACHED_SUCCESS);
 }
 
 Datum memcache_flush_all0(PG_FUNCTION_ARGS)
@@ -342,35 +493,40 @@ Datum memcache_flush_all0(PG_FUNCTION_ARGS)
   memcached_return rc;
 
   rc = memcached_flush(globals.mc, opt_expire);
+  if (rc == MEMCACHED_BUFFERED)
+    {
+      globals.flush_needed = true;
+      PG_RETURN_NULL();
+    }
   if (rc != MEMCACHED_SUCCESS)
     elog(WARNING, "pgmemcache: memcached_flush: %s",
                   memcached_strerror(globals.mc, rc));
 
-  PG_RETURN_BOOL(rc == 0);
+  PG_RETURN_BOOL(rc == MEMCACHED_SUCCESS);
 }
 
 Datum memcache_get(PG_FUNCTION_ARGS)
 {
-  text *get_key, *ret;
-  char *string, *key;
-  size_t key_length, return_value_length;
+  text *ret;
+#ifdef USE_LIBMEMCACHED
+  char *string;
   uint32_t flags;
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+  const unsigned char *string;
+#endif /* USE_OMCACHE */
+  size_t return_value_length;
   memcached_return rc;
+  size_t key_length;
+  const char *key = get_arg_cstring(PG_GETARG_TEXT_P(0), &key_length, true);
 
-  if (PG_ARGISNULL(0))
-    elog(ERROR, "pgmemcache: key cannot be NULL");
-
-  get_key = PG_GETARG_TEXT_P(0);
-
-  key = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(get_key)));
-  key_length = strlen(key);
-
-  if (key_length < 1)
-    elog(ERROR, "pgmemcache: key cannot be an empty string");
-  if (key_length >= 250)
-    elog(ERROR, "pgmemcache: key too long");
-
+#ifdef USE_LIBMEMCACHED
   string = memcached_get(globals.mc, key, key_length, &return_value_length, &flags, &rc);
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+  rc = omcache_get(globals.mc, omc_cc_to_cuc(key), key_length, &string, &return_value_length,
+                   NULL, NULL, OMCACHE_READ_TIMEOUT);
+#endif /* USE_OMCACHE */
 
   if (rc != MEMCACHED_SUCCESS && rc != MEMCACHED_NOTFOUND)
     elog(ERROR, "pgmemcache: memcached_get: %s",
@@ -382,7 +538,9 @@ Datum memcache_get(PG_FUNCTION_ARGS)
   ret = (text *) palloc(return_value_length + VARHDRSZ);
   SET_VARSIZE(ret, return_value_length + VARHDRSZ);
   memcpy(VARDATA(ret), string, return_value_length);
+#ifdef USE_LIBMEMCACHED
   free(string);
+#endif /* USE_LIBMEMCACHED */
 
   PG_RETURN_TEXT_P(ret);
 }
@@ -392,25 +550,36 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
   ArrayType *array;
   int array_length, array_lbound, i;
   Oid element_type;
-  uint32_t flags;
   memcached_return rc;
   char typalign;
   int16 typlen;
   bool typbyval;
-  char **keys, *value;
-  size_t *key_lens, value_length;
+#ifdef USE_LIBMEMCACHED
+  uint32_t flags;
+  char *current_key, *current_val;
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+  const unsigned char *current_key, *current_val;
+#endif /* USE_OMCACHE */
+  size_t current_key_len, current_val_len;
   FuncCallContext *funcctx;
   MemoryContext oldcontext;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   struct internal_fctx {
-      char **keys;
       size_t *key_lens;
+#ifdef USE_LIBMEMCACHED
+      const char **keys;
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+      const unsigned char **keys;
+      omcache_req_t *requests;
+      size_t request_count;
+      omcache_value_t *values;
+      size_t value_count;
+#endif /* USE_OMCACHE */
   } *fctx;
 
-  if (PG_ARGISNULL(0))
-    elog(ERROR, "pgmemcache: get_multi key cannot be null");
-
   array = PG_GETARG_ARRAYTYPE_P(0);
   if (ARR_NDIM(array) != 1)
     elog(ERROR, "pgmemcache: only single dimension ARRAYs are supported, "
@@ -430,40 +599,53 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
         ereport(ERROR,
                 (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                  errmsg("function returning record called in context that cannot accept type record")));
-      fctx = (struct internal_fctx *) palloc(sizeof(*fctx));
-
       get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);
 
-      keys = palloc(sizeof(char *) * (array_length + 1 /* extra key for last memcached_fetch call */ ));
-      key_lens = palloc(sizeof(size_t) * (array_length + 1));
+      fctx = (struct internal_fctx *) palloc(sizeof(*fctx));
+      /* extra NULL key for last memcached_fetch call */
+      fctx->keys = palloc(sizeof(char *) * (array_length + 1));
+      fctx->key_lens = palloc(sizeof(size_t) * (array_length + 1));
+      fctx->keys[array_length] = 0;
+      fctx->key_lens[array_length] = 0;
 
-      /* initialize terminating extra-key */
-      keys[array_length] = 0;
-      key_lens[array_length] = 0;
+#ifdef USE_LIBMEMCACHED
+      for (i = 0; i < array_length; i++)
+        {
+          int offset = array_lbound + i;
+          bool isnull;
+          Datum elem = array_ref(array, 1, &offset, 0, typlen, typbyval, typalign, &isnull);
+          if (!isnull)
+            fctx->keys[i] = get_arg_cstring(DatumGetTextP(elem), &fctx->key_lens[i], true);
+        }
+
+      rc = memcached_mget(globals.mc, fctx->keys, fctx->key_lens, array_length);
+      if (rc != MEMCACHED_SUCCESS)
+        elog(ERROR, "pgmemcache: memcached_mget: %s",
+                    memcached_strerror(globals.mc, rc));
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+      /* persistent request structures to handle pending requests */
+      fctx->requests = palloc(sizeof(omcache_req_t) * array_length);
+      fctx->request_count = array_length;
+      fctx->values = palloc(sizeof(omcache_value_t) * array_length);
+      fctx->value_count = array_length;
 
       for (i = 0; i < array_length; i++)
         {
           int offset = array_lbound + i;
           bool isnull;
-          Datum elem;
-
-          elem = array_ref(array, 1, &offset, 0, typlen, typbyval, typalign, &isnull);
+          Datum elem = array_ref(array, 1, &offset, 0, typlen, typbyval, typalign, &isnull);
           if (!isnull)
-            {
-              keys[i] = TextDatumGetCString(PointerGetDatum(elem));
-              key_lens[i] = strlen(keys[i]);
-            }
+            fctx->keys[i] = (const unsigned char *)
+              get_arg_cstring(DatumGetTextP(elem), &fctx->key_lens[i], true);
         }
-      fctx->keys = keys;
-      fctx->key_lens = key_lens;
 
-      rc = memcached_mget(globals.mc, (const char **) keys, key_lens, array_length);
-      if (rc != MEMCACHED_SUCCESS)
-        elog(ERROR, "pgmemcache: memcached_mget: %s",
-                    memcached_strerror(globals.mc, rc));
-
-      if (rc == MEMCACHED_NOTFOUND)
-        PG_RETURN_NULL();
+      rc = omcache_get_multi(globals.mc, fctx->keys, fctx->key_lens, array_length,
+                             fctx->requests, &fctx->request_count, fctx->values, &fctx->value_count,
+                             OMCACHE_READ_TIMEOUT);
+      if (rc != OMCACHE_OK && rc != OMCACHE_AGAIN)
+        elog(ERROR, "pgmemcache: omcache_get_multi: %s", omcache_strerror(rc));
+#endif /* USE_OMCACHE */
 
       attinmeta = TupleDescGetAttInMetadata(tupdesc);
       funcctx->attinmeta = attinmeta;
@@ -475,35 +657,60 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
   fctx = funcctx->user_fctx;
   attinmeta = funcctx->attinmeta;
 
-  value = memcached_fetch(globals.mc, fctx->keys[funcctx->call_cntr], &fctx->key_lens[funcctx->call_cntr], &value_length, &flags, &rc);
-  if (value != NULL)
+#ifdef USE_LIBMEMCACHED
+  current_key = (char *) fctx->keys[funcctx->call_cntr];  // cast away constness for libmemcached api
+  current_key_len = fctx->key_lens[funcctx->call_cntr];
+  current_val = memcached_fetch(globals.mc, current_key, &current_key_len, &current_val_len, &flags, &rc);
+  if (rc == MEMCACHED_END)
+    {
+      SRF_RETURN_DONE(funcctx);
+    }
+  else if (rc != MEMCACHED_SUCCESS)
+    {
+      elog(ERROR, "pgmemcache: memcached_fetch: %s",
+                  memcached_strerror(globals.mc, rc));
+      SRF_RETURN_DONE(funcctx);
+    }
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+  current_val = NULL;
+  if (fctx->value_count == 0 && fctx->request_count > 0)
+    {
+      fctx->value_count = fctx->request_count;
+      rc = omcache_io(globals.mc, fctx->requests, &fctx->request_count,
+                      fctx->values, &fctx->value_count, OMCACHE_READ_TIMEOUT);
+      if (rc != OMCACHE_OK && rc != OMCACHE_AGAIN)
+        elog(ERROR, "pgmemcache: omcache_io: %s", omcache_strerror(rc));
+    }
+  if (fctx->value_count > 0)
+    {
+      fctx->value_count --;
+      current_key = fctx->values[fctx->value_count].key;
+      current_key_len = fctx->values[fctx->value_count].key_len;
+      current_val = fctx->values[fctx->value_count].data;
+      current_val_len = fctx->values[fctx->value_count].data_len;
+    }
+#endif /* USE_OMCACHE */
+  if (current_val != NULL)
     {
       char **values;
       HeapTuple tuple;
       Datum result;
 
-      if (rc == MEMCACHED_END)
-        {
-          SRF_RETURN_DONE(funcctx);
-        }
-      else if (rc != MEMCACHED_SUCCESS)
-        {
-          elog(ERROR, "pgmemcache: memcached_fetch: %s",
-                      memcached_strerror(globals.mc, rc));
-          SRF_RETURN_DONE(funcctx);
-        }
       values = (char **) palloc(2 * sizeof(char *));
       /* make sure we have space for terminating zero character */
-      values[0] = (char *) palloc(fctx->key_lens[funcctx->call_cntr] + 1);
-      values[1] = (char *) palloc(value_length + 1);
+      values[0] = (char *) palloc(current_key_len + 1);
+      values[1] = (char *) palloc(current_val_len + 1);
 
-      memcpy(values[0], fctx->keys[funcctx->call_cntr], fctx->key_lens[funcctx->call_cntr]);
-      memcpy(values[1], value, value_length);
-      free(value);
+      memcpy(values[0], current_key, current_key_len);
+      memcpy(values[1], current_val, current_val_len);
+#ifdef USE_LIBMEMCACHED
+      free(current_val);
+#endif /* USE_LIBMEMCACHED */
 
       /* BuildTupleFromCStrings needs correct zero-terminated C-string, so terminate our raw strings */
-      values[0][fctx->key_lens[funcctx->call_cntr]] = '\0';
-      values[1][value_length] = '\0';
+      values[0][current_key_len] = '\0';
+      values[1][current_val_len] = '\0';
 
       tuple = BuildTupleFromCStrings(attinmeta, values);
       result = HeapTupleGetDatum(tuple);
@@ -513,76 +720,54 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
   SRF_RETURN_DONE(funcctx);
 }
 
-Datum memcache_incr(PG_FUNCTION_ARGS)
-{
-  return memcache_atomic_op(true, fcinfo);
-}
-
 Datum memcache_replace(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_REPLACE | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_REPLACE | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
 }
 
 Datum memcache_replace_absexpire(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_REPLACE | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_REPLACE | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
 }
 
 Datum memcache_set(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_SET | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_SET | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
 }
 
 Datum memcache_set_absexpire(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_SET | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_SET | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
 }
 
 Datum memcache_prepend(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_PREPEND | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_PREPEND | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
 }
 
 Datum memcache_prepend_absexpire(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_PREPEND | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_PREPEND | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
 }
 
 Datum memcache_append(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_APPEND | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_APPEND | PG_MEMCACHE_TYPE_INTERVAL, fcinfo);
 }
 
 Datum memcache_append_absexpire(PG_FUNCTION_ARGS)
 {
-  return memcache_set_cmd(PG_MEMCACHE_APPEND | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
+  return memcache_set_cmd(PG_MEMCACHE_CMD_APPEND | PG_MEMCACHE_TYPE_TIMESTAMP, fcinfo);
 }
 
 static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
 {
   memcached_return rc = MEMCACHED_FAILURE;
   const char *func = NULL;
-  char *key, *value;
-  text *key_text = NULL, *value_text;
-  size_t key_length, value_length;
   time_t expiration = 0;
-
-  if (PG_ARGISNULL(0))
-    elog(ERROR, "pgmemcache: key cannot be NULL");
-  if (PG_ARGISNULL(1))
-    elog(ERROR, "pgmemcache: value cannot be NULL");
-
-  key_text = PG_GETARG_TEXT_P(0);
-  key_length = VARSIZE(key_text) - VARHDRSZ;
-
-  /* These aren't really needed as we set libmemcached behavior to check for all invalid sets */
-  if (key_length < 1)
-    elog(ERROR, "pgmemcache: key cannot be an empty string");
-  if (key_length >= 250)
-    elog(ERROR, "pgmemcache: key too long");
-
-  value_text = PG_GETARG_TEXT_P(1);
-  value_length = VARSIZE(value_text) - VARHDRSZ;
+  size_t key_length, value_length;
+  const char *key = get_arg_cstring(PG_GETARG_TEXT_P(0), &key_length, true);
+  const char *value = get_arg_cstring(PG_GETARG_TEXT_P(1), &value_length, false);
 
   if (PG_NARGS() >= 3 && PG_ARGISNULL(2) == false)
     {
@@ -617,40 +802,37 @@ static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
         }
     }
 
-  key = VARDATA(key_text);
-  value = VARDATA(value_text);
-
-  if (type & PG_MEMCACHE_ADD)
+  switch (type & PG_MEMCACHE_CMD_MASK)
     {
+    case PG_MEMCACHE_CMD_ADD:
       func = "memcached_add";
       rc = memcached_add(globals.mc, key, key_length, value, value_length, expiration, 0);
-    }
-  else if (type & PG_MEMCACHE_REPLACE)
-    {
+      break;
+    case PG_MEMCACHE_CMD_REPLACE:
       func = "memcached_replace";
       rc = memcached_replace(globals.mc, key, key_length, value, value_length, expiration, 0);
-    }
-  else if (type & PG_MEMCACHE_SET)
-    {
+      break;
+    case PG_MEMCACHE_CMD_SET:
       func = "memcached_set";
       rc = memcached_set(globals.mc, key, key_length, value, value_length, expiration, 0);
-    }
-  else if (type & PG_MEMCACHE_PREPEND)
-    {
+      break;
+    case PG_MEMCACHE_CMD_PREPEND:
       func = "memcached_prepend";
       rc = memcached_prepend(globals.mc, key, key_length, value, value_length, expiration, 0);
-    }
-  else if (type & PG_MEMCACHE_APPEND)
-    {
+      break;
+    case PG_MEMCACHE_CMD_APPEND:
       func = "memcached_append";
       rc = memcached_append(globals.mc, key, key_length, value, value_length, expiration, 0);
-    }
-  else
-    {
+      break;
+    default:
       elog(ERROR, "pgmemcache: unknown set command type: %d", type);
-      return false;
     }
 
+  if (rc == MEMCACHED_BUFFERED)
+    {
+      globals.flush_needed = true;
+      PG_RETURN_NULL();
+    }
   if (rc != MEMCACHED_SUCCESS)
     elog(WARNING, "pgmemcache: %s: %s", func,
                   memcached_strerror(globals.mc, rc));
@@ -660,13 +842,10 @@ static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
 
 Datum memcache_server_add(PG_FUNCTION_ARGS)
 {
-  text *server = PG_GETARG_TEXT_P(0);
-  char *host_str;
-  memcached_return rc;
-
-  host_str = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(server)));
-
-  rc = do_server_add(host_str);
+  size_t host_len;
+  const char *host_buf = get_arg_cstring(PG_GETARG_TEXT_P(0), &host_len, false);
+  char *host = pnstrdup(host_buf, host_len);
+  memcached_return rc = do_server_add(host);
   if (rc != MEMCACHED_SUCCESS)
     elog(WARNING, "pgmemcache: memcached_server_push: %s",
                   memcached_strerror(globals.mc, rc));
@@ -686,10 +865,18 @@ static memcached_return do_server_add(const char *host_str)
   return rc;
 }
 
+#ifdef USE_LIBMEMCACHED
+#define MC_ENUM_INVAL -1
 #define MC_STR_TO_ENUM(d,v) \
-  if (strncmp(value, "MEMCACHED_" #d "_" #v, 9 + sizeof(#d) + sizeof(#v)) == 0 || \
-      strncmp(value, #v, sizeof(#v) - 1) == 0) \
+  if (strcmp(value, "MEMCACHED_" #d "_" #v) == 0 || strcmp(value, #v) == 0) \
       return MEMCACHED_##d##_##v
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+#define MC_ENUM_INVAL NULL
+#define MC_STR_TO_ENUM(d,v) \
+  if (strcmp(value, "MEMCACHED_" #d "_" #v) == 0 || strcmp(value, #v) == 0) \
+      return #v
+#endif /* USE_OMCACHE */
 
 static memcached_behavior get_memcached_behavior_flag(const char *value)
 {
@@ -708,6 +895,9 @@ static memcached_behavior get_memcached_behavior_flag(const char *value)
   MC_STR_TO_ENUM(BEHAVIOR, IO_MSG_WATERMARK);
   MC_STR_TO_ENUM(BEHAVIOR, KETAMA);
   MC_STR_TO_ENUM(BEHAVIOR, KETAMA_HASH);
+#ifdef USE_OMCACHE
+  MC_STR_TO_ENUM(BEHAVIOR, KETAMA_PRE1010);
+#endif
   MC_STR_TO_ENUM(BEHAVIOR, KETAMA_WEIGHTED);
   MC_STR_TO_ENUM(BEHAVIOR, NO_BLOCK);
   MC_STR_TO_ENUM(BEHAVIOR, NOREPLY);
@@ -731,31 +921,10 @@ static memcached_behavior get_memcached_behavior_flag(const char *value)
   MC_STR_TO_ENUM(BEHAVIOR, VERIFY_KEY);
 
   elog(ERROR, "pgmemcache: unknown behavior flag: %s", value);
-  return -1;
+  return MC_ENUM_INVAL;
 }
 
-static uint64_t get_memcached_behavior_data (const char *flag, const char *data)
-{
-  char *endptr;
-  uint64_t ret;
-
-  switch (get_memcached_behavior_flag(flag))
-    {
-    case MEMCACHED_BEHAVIOR_HASH:
-    case MEMCACHED_BEHAVIOR_KETAMA_HASH:
-      return get_memcached_hash_type(data);
-    case MEMCACHED_BEHAVIOR_DISTRIBUTION:
-      return get_memcached_distribution_type(data);
-    default:
-      ret = strtol(data, &endptr, 10);
-      if (endptr == data)
-        elog(ERROR, "pgmemcache: invalid behavior param %s: %s", flag, data);
-      return ret;
-    }
-  return ret;
-}
-
-static uint64_t get_memcached_hash_type(const char *value)
+static memcached_hash get_memcached_hash_type(const char *value)
 {
   MC_STR_TO_ENUM(HASH, MURMUR);
   MC_STR_TO_ENUM(HASH, MD5);
@@ -769,10 +938,10 @@ static uint64_t get_memcached_hash_type(const char *value)
   MC_STR_TO_ENUM(HASH, CRC);
 
   elog(ERROR, "pgmemcache: invalid hash name: %s", value);
-  return 0xffffffff; /* to avoid warning */
+  return MC_ENUM_INVAL;
 }
 
-static uint64_t get_memcached_distribution_type(const char *value)
+static memcached_server_distribution get_memcached_distribution_type(const char *value)
 {
   MC_STR_TO_ENUM(DISTRIBUTION, RANDOM);
   MC_STR_TO_ENUM(DISTRIBUTION, MODULA);
@@ -780,53 +949,119 @@ static uint64_t get_memcached_distribution_type(const char *value)
   MC_STR_TO_ENUM(DISTRIBUTION, CONSISTENT);
 
   elog(ERROR, "pgmemcache: invalid distribution name: %s", value);
-  return 0xffffffff; /* to avoid warning */
+  return MC_ENUM_INVAL;
+}
+
+static uint64_t get_memcached_behavior_data(const char *flag, const char *data, const char **val)
+{
+  char *endptr;
+  uint64_t ret;
+  memcached_behavior bkey = get_memcached_behavior_flag(flag);
+
+#ifdef USE_LIBMEMCACHED
+  switch (bkey)
+    {
+    case MEMCACHED_BEHAVIOR_HASH:
+    case MEMCACHED_BEHAVIOR_KETAMA_HASH:
+      return get_memcached_hash_type(data);
+    case MEMCACHED_BEHAVIOR_DISTRIBUTION:
+      return get_memcached_distribution_type(data);
+    default:
+#endif /* USE_LIBMEMCACHED */
+#ifdef USE_OMCACHE
+  ret = 0;
+  if (strcmp(bkey, "HASH") == 0 || strcmp(bkey, "KETAMA_HASH") == 0)
+    *val = get_memcached_hash_type(data);
+  else if (strcmp(bkey, "DISTRIBUTION") == 0)
+    *val = get_memcached_distribution_type(data);
+  else
+    {
+#endif /* USE_OMCACHE */
+      ret = strtol(data, &endptr, 10);
+      if (endptr == data)
+        elog(ERROR, "pgmemcache: invalid behavior param %s: %s", flag, data);
+    }
+  return ret;
 }
 
 /* NOTE: memcached_server_fn specifies that the first argument is const, but
  * memcached_stat_get_keys wants a non-const argument so we don't define it
  * as const here.
  */
-static memcached_return_t server_stat_function(memcached_st *ptr,
+static memcached_return_t server_stat_function(memcached_st *mc,
                                                memcached_server_instance_st server,
                                                void *context)
 {
-  char **list, **stat_ptr;
   memcached_return rc;
+  StringInfoData *strbuf = (StringInfoData *) context;
+  const char *hostname = memcached_server_name(server);
+  unsigned int port = memcached_server_port(server);
+
+  appendStringInfo(strbuf, "Server: %s (%u)\n", hostname, port);
+
+  {
+#ifdef USE_LIBMEMCACHED
   memcached_stat_st stat;
+  char **list, **stat_ptr;
 
-  rc = memcached_stat_servername(&stat, NULL, memcached_server_name(server),
-                                 memcached_server_port(server));
+  rc = memcached_stat_servername(&stat, NULL, hostname, port);
+  if (rc != MEMCACHED_SUCCESS)
+    return rc;
 
-  list = memcached_stat_get_keys(ptr, &stat, &rc);
+  list = memcached_stat_get_keys(mc, &stat, &rc);
+  if (rc != MEMCACHED_SUCCESS)
+    return rc;
 
-  appendStringInfo(context, "Server: %s (%u)\n", memcached_server_name(server),
-                   memcached_server_port(server));
-  for (stat_ptr = list; *stat_ptr; stat_ptr++)
+  for (stat_ptr = list; stat_ptr && *stat_ptr; stat_ptr++)
     {
-      char *value = memcached_stat_get_value(ptr, &stat, *stat_ptr, &rc);
-      appendStringInfo(context, "%s: %s\n", *stat_ptr, value);
+      char *value = memcached_stat_get_value(mc, &stat, *stat_ptr, &rc);
+      appendStringInfo(strbuf, "%s: %s\n", *stat_ptr, value);
       free(value);
     }
-
   free(list);
+#endif /* USE_LIBMEMCACHED */
+
+#ifdef USE_OMCACHE
+  size_t i, value_count = 50;
+  omcache_value_t values[50];
+  rc = omcache_stat(globals.mc, NULL, values, &value_count,
+                    server->server_index, OMCACHE_READ_TIMEOUT);
+  if (rc != OMCACHE_OK)
+    {
+      value_count = 0;
+      appendStringInfo(strbuf, "omcache_stat failed: %s\n", omcache_strerror(rc));
+    }
+
+  for (i = 0; i < value_count; i++)
+    {
+      int key_len = (int) values[i].key_len,
+          data_len = (int) values[i].data_len;
+      if (key_len == 0 && data_len == 0)
+        break;
+      appendStringInfo(strbuf, "%.*s: %.*s\n",
+                       key_len, (const char *) values[i].key,
+                       data_len, (const char *) values[i].data);
+    }
+#endif /* USE_OMCACHE */
+  }
+
+  appendStringInfo(strbuf, "\n");
   return MEMCACHED_SUCCESS;
 }
 
 Datum memcache_stats(PG_FUNCTION_ARGS)
 {
-  StringInfoData buf;
+  StringInfoData strbuf;
   memcached_return rc;
   memcached_server_fn callbacks[1];
 
-  initStringInfo(&buf);
+  initStringInfo(&strbuf);
   callbacks[0] = (memcached_server_fn) server_stat_function;
-  appendStringInfo(&buf, "\n");
-  rc = memcached_server_cursor(globals.mc, callbacks, (void *) &buf, 1);
+  rc = memcached_server_cursor(globals.mc, callbacks, (void *) &strbuf, 1);
 
   if (rc != MEMCACHED_SUCCESS && rc != MEMCACHED_SOME_ERRORS)
     elog(WARNING, "pgmemcache: memcache_stats: %s",
                   memcached_strerror(globals.mc, rc));
 
-  PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(buf.data)));
+  PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(strbuf.data)));
 }
